#include <TFT_eSPI.h> 
#include <SPI.h>
#include <DHT.h>       
#include "Wire.h"
#include <RTClib.h>   
#include "NotoSansBold15.h"
#include "Roboto_Thin_24.h"
#include <Adafruit_BMP280.h>
#include <WiFi.h>
#include <ESPAsyncWebServer.h>
#include <vector>

// ============================DHT INIT==================================

#define DHTPIN 33         
#define DHTTYPE DHT11
DHT dht(DHTPIN, DHTTYPE);

// ============================CLOCK INIT=================================
#define CLOCK_SDA 32
#define CLOCK_SCL 13

TwoWire I2C_Clock = TwoWire(0);

RTC_DS3231 rtc;

// ============================DISPLAY INIT===============================

TFT_eSPI tft = TFT_eSPI();   
TFT_eSprite face = TFT_eSprite(&tft);
TFT_eSprite clock_string = TFT_eSprite(&tft);

// ===========================CLOCK ANIMATION=============================
#define CLOCK_X_POS 10
#define CLOCK_Y_POS 10

#define CLOCK_FG   0x0000
#define CLOCK_BG   0xdbb4
#define SECCOND_FG TFT_RED
#define LABEL_FG   TFT_GOLD

#define CLOCK_R       80.0f 
#define H_HAND_LENGTH CLOCK_R/2.0f
#define M_HAND_LENGTH CLOCK_R/1.4f
#define S_HAND_LENGTH CLOCK_R/1.3f

#define FACE_W CLOCK_R * 2 + 1
#define FACE_H CLOCK_R * 2 + 1

#define SECOND_ANGLE 360.0 / 60.0
#define MINUTE_ANGLE SECOND_ANGLE / 60.0
#define HOUR_ANGLE   MINUTE_ANGLE / 12.0

#define FACE_W CLOCK_R * 2 + 1
#define FACE_H CLOCK_R * 2 + 1

uint8_t h = 0, m = 0, s = 0;

float time_secs = h * 3600 + m * 60 + s;

uint32_t targetTime = 0;
// =========================AIR PRESSURE INIT===============================
#define BMP_SCK  25
#define BMP_MISO 14
#define BMP_MOSI 26
#define BMP_CS   27
Adafruit_BMP280 bmp(BMP_CS, BMP_MOSI, BMP_MISO, BMP_SCK);

// =========================================================================

#define TFT_GREY 0x5AEB
#define TFT_BACKGROUND 0x1003

#define LOOP_PERIOD 45 

uint32_t updateTime = 0;       

float lastTemperature = -999.0;
float lastHumidity = -999.0;

float lastPressure = -999.0;
float lastAltitude = -999.0;

// =========================WIFI INIT===============================
const char* ssid = "ada";
const char* password = "ko6zrg5ua1gft";

AsyncWebServer server(80);

String displayMode = "None";
String previousDisplayMode = "None";

uint16_t currentBgColor = TFT_BACKGROUND;
uint16_t previousBgColor = TFT_BACKGROUND;

// esp ip: 172.20.10.7

// =========================TIMER INIT===============================

hw_timer_t *timer = NULL;
volatile bool changeFace = false;

void IRAM_ATTR timerFunction() {
    changeFace = !changeFace;
}

// =========================ALARMS/NOTIFICATIONS=======================

struct Alarm {
  DateTime time;
  String message;

  bool operator==(const Alarm& other) const {
    return time.unixtime() == other.time.unixtime() && message == other.message;
  }
};

std::vector<Alarm> alarms;

bool alarmIsActive = false;

String alarmMessage = "";

// =====================================================================


void setup(void) {
  Serial.begin(921600);
  delay(100); 

  /* WIFI */
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi...");
  }

  Serial.println(WiFi.localIP());

  server.on("/display", HTTP_GET, [](AsyncWebServerRequest *request){
  if (request->hasParam("mode")) {
      displayMode = request->getParam("mode")->value();
      Serial.println(displayMode);
    }

    if (request->hasParam("time") && request->hasParam("message")) {
      Serial.println("alarm");
      String alarm_time = request->getParam("time")->value();
      Serial.println(alarm_time);
      String alarm_message = request->getParam("message")->value();

      DateTime alarmTime = parseTimeStringRequest(alarm_time);
      alarms.push_back({alarmTime, alarm_message});
    }

    if (request->hasParam("alarm") && request->getParam("alarm")->value() == "stop") {
      // change back to the previous display mode
      displayMode = previousDisplayMode;
      alarmIsActive = false;
    }
    request->send(200, "text/plain", "Display mode set to: " + displayMode);
  });
  
  server.begin();

  /* CLOCK */

  I2C_Clock.begin(CLOCK_SDA, CLOCK_SCL);
  bool status;
  status = rtc.begin(&I2C_Clock);

  /* AIR PRESSURE */

  status = bmp.begin(0x76);
  bmp.setSampling(Adafruit_BMP280::MODE_NORMAL,
                  Adafruit_BMP280::SAMPLING_X2, 
                  Adafruit_BMP280::SAMPLING_X16,  
                  Adafruit_BMP280::FILTER_X16,      
                  Adafruit_BMP280::STANDBY_MS_500);

  /* TEMP */

  dht.begin();

  /* DISPLAY */

  tft.init();
  tft.setRotation(3);

  tft.fillScreen(TFT_BACKGROUND);

  tft.setTextSize(3);  
  tft.setTextColor(TFT_WHITE); 

  if (rtc.lostPower()) {
    rtc.adjust(DateTime(F(__DATE__), F(__TIME__)));
  }

  /* initialize things for clock animation */
  face.createSprite(FACE_W, FACE_H);
  face.loadFont(NotoSansBold15);

  /* TIMER */

  timer = timerBegin(0, 80, true);
  timerAttachInterrupt(timer, &timerFunction, true);
  timerAlarmWrite(timer, 10000000, true);
  timerAlarmEnable(timer);
  updateTime = millis() + 100;

}

void loop() {

  if (updateTime <= millis()) {
    float temperature = dht.readTemperature();
    float humidity = dht.readHumidity();

    if (displayMode == "time1") {
      currentBgColor = tft.color565(128, 0, 128);
    } else if (displayMode == "time2" || displayMode == "temp1" || displayMode == "temp2" || displayMode == "hum1"
                || displayMode == "hum2" || displayMode == "default1" || displayMode == "default2") {
      currentBgColor = TFT_BACKGROUND;
    } else if (displayMode == "press1") {
      currentBgColor = 0x0aaa;
    } else if (displayMode == "press2") {
      currentBgColor = 0x8141;
    } else if (displayMode == "alarm") {
      currentBgColor = 0xe57f;
    } else {
      currentBgColor = TFT_BACKGROUND;
    }

    if (displayMode != previousDisplayMode || currentBgColor != previousBgColor) {
      tft.fillScreen(currentBgColor);
      previousDisplayMode = displayMode;
      previousBgColor = currentBgColor;
    }

    DateTime now = rtc.now();

    checkAlarms(now);

    // ################### DISPLAY MODES #####################
    
    if (displayMode == "time1" && !alarmIsActive) {
        showTime(now);

    } else if (displayMode == "time2" && !alarmIsActive) {
        renderFaceClock(now);

    } else if (displayMode == "temp1" && !alarmIsActive) {

      showTemperature(temperature);
      lastTemperature = temperature;

    } else if (displayMode == "temp2" && !alarmIsActive) {

      showTemperature2(temperature);
      lastTemperature = temperature;

    } else if (displayMode == "hum1" && !alarmIsActive) {
      showHumidity(humidity);

    } else if (displayMode == "hum2" && !alarmIsActive) {
      showHumidity2(humidity);

    } else if (displayMode == "press1" && !alarmIsActive) {
      float pressure = bmp.readPressure(); 
      float altitude = bmp.readAltitude(1020.33);
      showPressure(pressure, altitude);

    } else if (displayMode == "press2" && !alarmIsActive) {
      float pressure = bmp.readPressure(); 
      float altitude = bmp.readAltitude(1020.33);
      showPressure2(pressure, altitude);

    } else if (displayMode == "face1" && !alarmIsActive) {
      drawSmiley(tft.width() / 2, tft.height() / 2, 70, TFT_WHITE);

    } else if (displayMode == "face2" && !alarmIsActive) {
      drawSquareFace(tft.width() / 2, tft.height() / 2, 140, TFT_WHITE);

    } else if (!alarmIsActive){
      // default mode
      displayDefaultMode();
    } else if (alarmIsActive) {
      displayAlarm(alarmMessage);
    }

    // === UPDATE TIME
    updateTime = millis() + 100;
  }
  
}

// ============ ALARM ==============

// format: YYYY-MM-DD-HH-MM

DateTime parseTimeStringRequest(String timeString) {
  int year = timeString.substring(0, 4).toInt();
  int month = timeString.substring(5, 7).toInt();
  int day = timeString.substring(8, 10).toInt();
  int hour = timeString.substring(11, 13).toInt();
  int minute = timeString.substring(14, 16).toInt();
  return DateTime(year, month, day, hour, minute, 0);
}

void checkAlarms(DateTime now) {
  for (const auto& alarm : alarms) {
      if (alarm.time.unixtime() == now.unixtime()) {
        displayMode = "alarm";
        alarmIsActive = true;
        alarmMessage = alarm.message;
        displayAlarm(alarm.message);
        alarms.erase(std::remove(alarms.begin(), alarms.end(), alarm), alarms.end());
        break;
      }
  }
}

static int alarmBg = 0;

void displayAlarm(String message) {
    if (alarmBg == 0) {
      alarmBg = 1;
      tft.fillScreen(0xe57f);
    }

    tft.setTextSize(3);
    tft.setTextColor(TFT_BLACK, 0xe57f);

    int tX = 80;   
    int tY = 50;   
    int tW = 60; 
    int tH = 120;

    tft.fillTriangle(tX - tW / 2, tY, tX + tW / 2, tY, tX, tY + tH, TFT_RED);

    tft.fillCircle(tX, tY + tH + 25, 15, TFT_RED);

    tft.drawString(message, 150, 70);
    

}

// ============ DEFAULT DISPLAY MODE ==============

void displayDefaultMode() {
    if (changeFace == true) {
      drawSmiley(tft.width() / 2, tft.height() / 2, 70, TFT_WHITE);
      displayMode = "default1";
    } else if (changeFace == false) {
      drawSquareFace(tft.width() / 2, tft.height() / 2, 140, TFT_WHITE);
      displayMode = "default2";
    }
}

// ============ AIR PRESSURE ==============

static int pressureBg = 0;

void showPressure(float pressure, float altitude) {
  float pressure_hPa = pressure / 100.0;

  tft.drawRect(10, 10, tft.width() / 2 - 20, 10, TFT_WHITE);
  tft.drawRect(tft.width() / 2 + 10, 10, tft.width() / 2 - 20, 10, TFT_WHITE);

  tft.drawRect(10, tft.height() - 20, tft.width() / 2 - 20, 10, TFT_WHITE);
  tft.drawRect(tft.width() / 2 + 10, tft.height() - 20, tft.width() / 2 - 20, 10, TFT_WHITE);

  if (pressureBg == 0) {
    tft.fillScreen(0x0aaa);
    pressureBg = 1;
  }
  if (pressure != lastPressure) {
    lastPressure = pressure;
    tft.setTextSize(2);
    tft.setTextColor(0x8e99, 0x0aaa);
    tft.setCursor(15, 90);
    tft.print("Air pressure: ");
    tft.print(pressure_hPa);
    tft.println(" hPa");
  }

  if (altitude != lastAltitude) {
    lastAltitude = altitude;
    tft.setTextSize(2);
    tft.setTextColor(0x8e99, 0x0aaa);
    tft.setCursor(15, 120);
    tft.print("Altitude: ");
    tft.print(altitude);
    tft.println(" m");
  }
}

static int pressureBg2 = 0;

void showPressure2(float pressure, float altitude) {
  float pressure_hPa = pressure / 100.0;
  if (pressureBg2 == 0) {
    tft.fillScreen(0x8141);
    pressureBg2 = 1;
  }

  tft.drawLine(0, 20, tft.width(), 20, 0xea44);  
  tft.drawLine(0, 40, tft.width(), 40, 0xeca9);
  tft.drawLine(0, 60, tft.width(), 60, 0xee69);

  tft.drawLine(0, tft.height() - 60, tft.width(), tft.height() - 60, 0xea44);  
  tft.drawLine(0, tft.height() - 40, tft.width(), tft.height() - 40, 0xeca9);  
  tft.drawLine(0, tft.height() - 20, tft.width(), tft.height() - 20, 0xee69);

  if (pressure != lastPressure) {
    lastPressure = pressure;
    tft.setTextSize(2);
    tft.setTextColor(0xffba, 0x8141);
    tft.setCursor(15, 90);
    tft.print("Air pressure: ");
    tft.print(pressure_hPa);
    tft.println(" hPa");
  }

  if (altitude != lastAltitude) {
    lastAltitude = altitude;
    tft.setTextSize(2);
    tft.setTextColor(0xffba, 0x8141);
    tft.setCursor(15, 120);
    tft.print("Altitude: ");
    tft.print(altitude);
    tft.println(" m");
  }
}

// ============ TEMPERATURE ==============

void showTemperature(float temperature) {
  tft.setTextSize(3);     
  tft.setTextColor(0xDCBA, TFT_BACKGROUND); 
  tft.setCursor(50, 50);
  tft.println("Temperature:");

  tft.setTextSize(5);        
  tft.setTextColor(0xC49B, TFT_BACKGROUND); 
  tft.setCursor(50, 100);  
  tft.print(temperature);
  tft.println(" C");

  if (temperature > 20.00) {
    drawSun();
  } 
}

void drawSun() {
  
  int cX = tft.width() / 2;
  int cY = tft.height() - 40; 
  int size = 30;

  tft.fillCircle(cX, cY, size, TFT_YELLOW); 
  
  
  for (int angle = 0; angle < 360; angle += 45) {
    float x1 = cX + size * cos(angle * PI / 180.0);
    float y1 = cY + size * sin(angle * PI / 180.0);
    float x2 = cX + (size + 10) * cos(angle * PI / 180.0); 
    float y2 = cY + (size + 10) * sin(angle * PI / 180.0); 
    tft.drawLine(x1, y1, x2, y2, TFT_YELLOW);
  }

}

void showTemperature2(float temperature) {

  uint16_t colors[] = {TFT_RED, TFT_GREEN, TFT_BLUE, TFT_YELLOW, TFT_CYAN,
                       TFT_MAGENTA, TFT_ORANGE, TFT_WHITE, TFT_PINK, TFT_PURPLE};
  int x1 = 0, y1 = 0;
  int step = 5; 

  for (int i = 0; i < 10; i++) {
    int x2 = tft.width() - 1 - i * step;
    int y2 = tft.height() - 1 - i * step;
    tft.drawRect(x1, y1, x2 - x1, y2 - y1, colors[i % 10]);

    x1 += step;
    y1 += step;
  }

  tft.setTextSize(5);        
  tft.setTextColor(0xED7D, TFT_BACKGROUND); 
  tft.setCursor(55, 100);  
  tft.print(temperature);
  tft.println(" C");
}

// ============ HUMIDITY ==============

void showHumidity(float humidity) {
      tft.setTextSize(3);
      tft.setTextColor(TFT_WHITE, TFT_BACKGROUND);
      tft.setCursor(10, 100);
      tft.print("Humidity: ");
      tft.print(humidity);
      tft.println(" %");
      tft.drawCircle(20, 20, 20, 0xc90d);
      tft.drawCircle(40, 70, 20, 0x2499);
      tft.drawCircle(70, 150, 20, 0xc909);
      tft.drawCircle(140, 180, 20, 0x2677);
      tft.drawCircle(45, 210, 20, 0x2499);
      tft.drawCircle(135, 34, 20, 0x5919);
      tft.drawCircle(195, 55, 20, 0xb0af);
      tft.drawCircle(268, 28, 20, 0x2678);
      tft.drawCircle(200, 210, 20, 0x22f9);
      tft.drawCircle(285, 165, 20, 0xb919);
      lastHumidity = humidity;
}

void showHumidity2(float humidity) {
      tft.setTextColor(TFT_WHITE, TFT_BACKGROUND);
      tft.setCursor(110, 80);
      tft.setTextSize(2);
      tft.println("Humidity:");
      tft.setCursor(100, 120);
      tft.setTextSize(3);
      tft.print(humidity);
      tft.println(" %");
      tft.drawCircle(tft.width() / 2, tft.height() / 2, 100, TFT_PURPLE);
      tft.drawCircle(tft.width() / 2, tft.height() / 2, 90, TFT_PINK);
      lastHumidity = humidity;
}

// ============ FACES ==============

void drawSmiley(int x, int y, int radius, uint16_t color) {
  tft.fillCircle(x, y, radius, color); 

  int eyeRadius = radius / 8;
  int eyeOffset = radius / 3;
  tft.fillCircle(x - eyeOffset, y - eyeOffset, eyeRadius, TFT_BLACK); 
  tft.fillCircle(x + eyeOffset, y - eyeOffset, eyeRadius, TFT_BLACK);

  int smileRadius = radius / 2;
  int mouthY = y + radius / 4 - 5; 

  tft.drawArc(x, y + 10, smileRadius, smileRadius - 5, 270, 90, TFT_BLACK, TFT_WHITE);
}

void drawSquareFace(int x, int y, int size, uint16_t color) {

  tft.fillRect(x - size / 2, y - size / 2, size, size, color);
  
  int eyeSize = size / 8;
  int eyeX = size / 4;
  int eyeY = size / 4;

  tft.fillRect(x - eyeX - eyeSize / 2, y - eyeY - eyeSize / 2, eyeSize, eyeSize, TFT_BACKGROUND);

  tft.fillRect(x + eyeX - eyeSize / 2, y - eyeY - eyeSize / 2, eyeSize, eyeSize, TFT_BACKGROUND);

  int mW = size / 2;
  int mH = size / 8;
  tft.fillRect(x - mW / 2, y + size / 4 - mH / 2, mW, mH, TFT_BACKGROUND);
}

// ============ TIME ==============

static int timeFlag = 0;

void showTime(DateTime now) {
  uint32_t purpleColor = tft.color565(128, 0, 128);

  if (timeFlag == 0) {
    tft.fillScreen(purpleColor);
    timeFlag = 1;
  }

  static DateTime prevTime = now;

  if (now != prevTime) {
    tft.setTextColor(TFT_WHITE, purpleColor);
    tft.setTextSize(4);

    char buf2[] = "YYYY/MM/DD";
    tft.drawString(now.toString(buf2), 45, tft.height() / 2 - 45);

    if (now.hour() != prevTime.hour() || now.minute() != prevTime.minute() || now.second() != prevTime.second()) {
      char buf1[] = "hh:mm:ss";
      tft.drawString(now.toString(buf1), 70, tft.height() / 2 + 10);
    }
    
    prevTime = now;
  }
}

static int renderClockBg = 0;

static void renderFaceClock(DateTime t) {

  float h_angle = (t.hour() % 12 + t.minute() / 60.0) * 30.0;
  float m_angle = t.minute() * 6.0;
  float s_angle = t.second() * 6.0;

  face.fillSprite(TFT_BACKGROUND);

  int16_t centerX = face.width() / 2;
  int16_t centerY = face.height() / 2;
  int16_t clockRadius = CLOCK_R;       

  face.fillSmoothCircle(centerX, centerY, clockRadius, CLOCK_BG);

  face.setTextDatum(MC_DATUM);

  face.setTextColor(CLOCK_FG, CLOCK_BG);

  constexpr uint32_t dialOffset = CLOCK_R - 10;

  float xp = 0.0, yp = 0.0; 

  for (uint32_t h = 1; h <= 12; h++) {
    getCoord(centerX, centerY, &xp, &yp, clockRadius - 10, h * 30.0);
    face.drawNumber(h, xp, yp);
  }

  getCoord(centerX, centerY, &xp, &yp, M_HAND_LENGTH, m_angle);
  face.drawWideLine(centerX, centerY, xp, yp, 6.0f, CLOCK_FG);
  face.drawWideLine(centerX, centerY, xp, yp, 2.0f, CLOCK_BG);

  getCoord(centerX, centerY, &xp, &yp, H_HAND_LENGTH, h_angle);
  face.drawWideLine(centerX, centerY, xp, yp, 6.0f, CLOCK_FG);
  face.drawWideLine(centerX, centerY, xp, yp, 2.0f, CLOCK_BG);

  face.fillSmoothCircle(centerX, centerY, 4, CLOCK_FG);

  getCoord(centerX, centerY, &xp, &yp, S_HAND_LENGTH, s_angle);
  face.drawWedgeLine(centerX, centerY, xp, yp, 2.5, 1.0, SECCOND_FG);
  face.pushSprite(75, 40, TFT_TRANSPARENT);
}

#define DEG2RAD 0.0174532925
void getCoord(int16_t x, int16_t y, float *xp, float *yp, int16_t r, float a)
{
  float sx1 = cos( (a - 90) * DEG2RAD);
  float sy1 = sin( (a - 90) * DEG2RAD);
  *xp =  sx1 * r + x;
  *yp =  sy1 * r + y;
}



